import os
from typing import Union

import boto3
import httpx

from hangar_sdk.config import Config


class HangarScope:
    def __init__(
        self,
        region: str,
        name: str,
        accountName: str = "default",
        roleAssumptions: list = [],
        api_key: Union[str, None] = None,
        _entity: str = None,
        endpoint: str = "https://api.tryhangar.com/controlplane",
    ):
        if api_key is None:
            self.api_key = os.environ["HANGAR_API_KEY"]
        else:
            self.api_key = api_key

        if os.environ.get("HANGAR_URL") is not None:
            self.endpoint = os.environ["HANGAR_URL"]

        self.endpoint = endpoint
        self.region = region
        self.entity = _entity
        self.name = name
        # Maintained by api key
        self.roleAssumptions = roleAssumptions
        self.resources = {}
        self.accountName = accountName

    async def deploy(self):
        if Config.mode == "invoke":
            return
        try:
            all = self.get_resources()

        except Exception:
            all = []

        to_update = []
        to_add = []
        to_delete = []

        for control_plane_resource in all:
            if control_plane_resource in self.resources.keys():
                to_update.append(self.resources.pop(control_plane_resource))
            elif control_plane_resource not in self.resources.keys():
                to_delete.append(control_plane_resource)

        for local_resource in self.resources.values():
            to_add.append(local_resource)
        for resource in to_add + to_update:
            resource

        for resource in to_add + to_update:
            cr = resource._resolve()
            if cr is not None:
                await cr

        for resource in to_delete:
            self.delete_resource(resource_id=resource)

    def add_construct(self, data, force_change=False):
        payload = {
            "region": self.region,
            "version": "v1",
            "roleAssumptions": self.roleAssumptions,
            "tags": [],
            "scope": self.name,
            "accountName": self.accountName,
            "resourceId": data["resourceId"],
            **data,
            "forceChange": force_change,
        }

        # print(payload)
        try:
            if self.entity is not None:
                response = httpx.request(
                    "POST",
                    self.endpoint,
                    headers={"X-API-KEY": self.api_key},
                    json=payload,
                    timeout=20.0,
                )

            else:
                response = httpx.request(
                    "POST",
                    self.endpoint,
                    headers={"X-API-KEY": self.api_key},
                    json=payload,
                    timeout=20.0,
                )

        except httpx.ReadTimeout:
            raise Exception("Couldnt acquire lock")
        if response.status_code != 200:
            raise Exception(response.text)

        if "job_id" in response.json():
            return response.json()["job_id"]
        else:
            raise Exception(response.text)

    def get_logs(self, resource_id, path):
        res = httpx.request(
            "GET",
            self.endpoint + "/resources/" + resource_id + "/logs",
            headers={"X-API-KEY": self.api_key},
            params={"path": path},
        )

        return res.json()

    def get_state(self, resource_id):
        res = httpx.request(
            "GET",
            self.endpoint + "/state",
            headers={"X-API-KEY": self.api_key},
            params={
                "resourceId": resource_id,
            },
        )
        return res.json()

    def get_status(self, job_id):
        return httpx.request(
            "GET",
            self.endpoint + "/status",
            headers={"X-API-KEY": self.api_key},
            params={"job_id": job_id},
            timeout=httpx.Timeout(20.0),
        )

    def get_resources(self):
        r = httpx.request(
            "GET",
            f"{self.endpoint}/scopes/{self.name}/resources",
            headers={"X-API-KEY": self.api_key},
        )
        if "detail" in r.json():
            raise Exception(r.json()["detail"])

        return [resource["ResourceId"] for resource in r.json()]

    def execute_action(self, resource_id, action: str, params={}):
        payload = {
            "resourceId": resource_id,
            "action": action,
            "params": params,
        }

        if self.entity is not None:
            response = httpx.request(
                "PATCH",
                self.endpoint,
                headers={"X-API-KEY": self.api_key},
                json=payload,
                params={"entityId": self.entity},
                timeout=20.0,
            )

        else:
            response = httpx.request(
                "PATCH",
                self.endpoint,
                headers={"X-API-KEY": self.api_key},
                json=payload,
                timeout=20.0,
            )

        if response.status_code != 200:
            raise Exception(response.text)

        return response.json()

    def delete_resource(self, resource_id):
        print("Deleting resource: " + resource_id)
        return httpx.request(
            "DELETE",
            self.endpoint,
            headers={"X-API-KEY": self.api_key},
            params={"resourceId": resource_id},
        )

    def get_resource(self, resource_id):
        return httpx.request(
            "GET",
            self.endpoint + "/resources/" + resource_id,
            headers={"X-API-KEY": self.api_key},
        ).json()

    def get_boto3_session(self):
        credentials = httpx.request(
            "GET",
            self.endpoint + "/scopes/" + self.name + "/credentials",
            headers={"X-API-KEY": self.api_key},
            timeout=30.0,
        ).json()
        return boto3.Session(
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
            region_name=self.region,
        )



from .base import HangarScope
from .constructs import (
    S3,
    Asset,
    BuildkitBuilder,
    Cluster,
    Container,
    ContainerBuilder,
    DirPath,
    GitHubRepo,
    PortMappings,
    Registry,
    Service,
)
from .main import app

__all__ = [
    "HangarScope",
    "Asset",
    "BuildkitBuilder",
    "Cluster",
    "Container",
    "ContainerBuilder",
    "GitHubRepo",
    "PortMappings",
    "Registry",
    "Service",
    "DirPath",
    "S3",
    "app",
]



import asyncio
from abc import ABC, abstractmethod
from typing import List, Optional
from attr import define
import attr
from attrs import field

from rich import print
from rich.live import Live
from rich.tree import Tree

from hangar_sdk.config import Config

from .base import HangarScope
from .utils import print_status


class HangarManager:
    objects = []

    @staticmethod
    def add(obj):
        HangarManager.objects.append(obj)

    @staticmethod
    def delete_objects():
        while HangarManager.objects:
            obj = HangarManager.objects.pop()
            del obj
            print("Object deleted.")


@attr.s(kw_only=True, slots=False)
class Resource(ABC):
    name: str
    _resolved: bool = False

    def __attrs_post_init__(self):
        HangarManager.add(self)
        self.mode = Config.mode

    async def deploy(self):
        await self._resolve()

    def _resolve(self, parent=None):
        pass


class Deployable:
    @abstractmethod
    def create_construct_definition(self) -> dict:
        pass


@define(kw_only=True, slots=False)
class CompositeResource(Resource, ABC):
    name: str
    scope: HangarScope
    _changed: bool = False
    _job_id: Optional[str] = None
    _resources: List[Resource] = field(factory=list)
    _dependencies: List[Resource] = field(factory=list)
    _parents: List[Resource] = field(factory=list)
    mode: str = Config.mode
    _numbers: List[int] = field(factory=list)
    _resolved: bool = False

    def __attrs_post_init__(self):
        HangarManager.add(self)
        self.mode = Config.mode
        self.scope.resources[self.name] = self
        self._numbers.append(1)

        for k, f in self.__dict__.items():
            if k.startswith("_"):
                continue
            if isinstance(f, CompositeResource):
                self._depends_on(f)

            elif isinstance(f, list):
                for item in f:
                    # pass
                    # # print(item)
                    if isinstance(item, CompositeResource):
                        self._depends_on(item)

    def __post_init__(self, *args, **kwargs):
        self.__attrs_post_init__(*args, **kwargs)

    def _depended_on_by(self, resource: Resource):
        resource._dependencies.append(self)
        self._parents.append(resource)

    def _depends_on(self, resource: Resource):
        # if resource in self._dependencies:
        #     return
        self._dependencies.append(resource)
        resource._parents.append(self)

    def _get_ref(self):
        return {"!REF": True, "resourceId": self.name}

    def pre_resolve(self):
        pass

    def post_resolve(self):
        pass

    async def _resolveDependencies(self):
        for dependency in self._dependencies:
            await dependency._resolve(self)

    async def _resolveDependedOnBy(self):
        for dependency in self._parents:
            await dependency._resolve(self)

    def _create_full_construct_definition(self):
        return {
            "config": {**self.create_construct_definition(), "name": self.name},
            "resourceId": self.name,
            "parents": [parent.name for parent in self._parents],
        }

    async def _resolve(self, parent: Optional[Resource] = None):
        # assert 6 == 9

        if self._resolved:
            return

        if self.mode == "delete":
            await self._resolveDependedOnBy()
            self.scope.delete_resource(self.name)
            self._resolved = True
            return
        if self.mode == "invoke":
            return
        await self._resolveDependencies()

        to_change = False
        for dependency in self._dependencies:
            if dependency._changed:
                to_change = True

        self.pre_resolve()

        if not self._resolved:
            if isinstance(self, Deployable):
                job_id = self.scope.add_construct(
                    self._create_full_construct_definition(),
                    force_change=to_change,
                )

                self._job_id = job_id

                await self.poll_status()

            self._resolved = True

        self.post_resolve()

    def delete(self):
        self.scope.delete_resource(self.name)

    async def poll_status(self):
        if self._job_id is None:
            raise Exception("No job id")

        response = None

        with Live(print_status(self.name, None)[0], refresh_per_second=4) as live:
            while True:
                if self._job_id == "no-change":
                    tree = Tree(self.name)
                    tree.add("No change")
                    live.update(tree)
                    self._changed = False
                    break
                self._changed = True

                response = self.scope.get_status(self._job_id)
                await asyncio.sleep(1)

                if response.status_code != 200:
                    raise Exception(response.text)

                status = response.json()
                render, _, errors = print_status(self.name, status)
                live.update(render)

                if errors and len(errors) > 0:
                    raise Exception(f"Errors occurred: {errors}")

                if response.json() and response.json()["completed"] is True:
                    break

        return response.json() if response else None

    @property
    def state(self):
        return self.scope.get_state(self.name)



from rich.tree import Tree


def print_status(resource, status):
    if status is None:
        return Tree(resource + " : " + "Pending"), None, None
    # print(status)
    items = list(
        map(
            lambda x: {
                "resource_id": x["resource_id"],
                "resource_type": x["resource_type"],
                "status": x["status"],
            },
            list(status["resources"].values()),
        )
    )
    items = sorted(items, key=lambda x: x["resource_type"])
    errors = list(
        map(
            lambda x: {"error_type": "InternalHangarError", "message": x}
            if isinstance(x, str)
            else {
                "resource_id": x["resource_type"],
                "resource_type": x["resource_id"],
                "message": x["message"].split("\n")[0],
            },
            list(status["errors"].values()),
        )
    )
    # print(errors)

    tree = Tree(resource)
    for v in items:
        rendered_status = (
            v["resource_type"] + " - " + v["resource_id"] + " " + v["status"]
        )
        matches = [
            x
            for x in errors
            if (
                "resource_type" in x
                and (
                    v["resource_type"] == x["resource_type"]
                    and v["resource_id"] == x["resource_id"]
                )
            )
            or "error_type" in x
        ]
        # print(matches)

        if len(matches) != 0:
            rendered_status += " error: " + matches[0]["message"]

        tree.add(rendered_status)

    return tree, items, errors



import contextlib
import io
import json
import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import zipfile
from collections import OrderedDict  # noqa
from datetime import datetime, timedelta
from os import PathLike  # noqa
from typing import cast  # noqa
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Iterator,
    List,  # noqa
    MutableMapping,
    Optional,
    Tuple,
    Union,
)

import click
import dateutil.parser
from dateutil.tz import tzutc

WELCOME_PROMPT = "hi"

OptInt = Optional[int]
OptBytes = Optional[bytes]
EnvVars = MutableMapping
StrPath = Union[str, "PathLike[str]"]


class AbortedError(Exception):
    pass


def to_cfn_resource_name(name: str) -> str:
    """Transform a name to a valid cfn name.

    This will convert the provided name to a CamelCase name.
    It's possible that the conversion to a CFN resource name
    can result in name collisions.  It's up to the caller
    to handle name collisions appropriately.

    """
    if not name:
        raise ValueError("Invalid name: %r" % name)
    word_separators = ["-", "_"]
    for word_separator in word_separators:
        word_parts = [p for p in name.split(word_separator) if p]
        name = "".join([w[0].upper() + w[1:] for w in word_parts])
    return re.sub(r"[^A-Za-z0-9]+", "", name)


def remove_stage_from_deployed_values(key: str, filename: str) -> None:
    """Delete a top level key from the deployed JSON file."""
    final_values: Dict[str, Any] = {}
    try:
        with open(filename, "r") as f:
            final_values = json.load(f)
    except IOError:
        # If there is no file to delete from, then this funciton is a noop.
        return

    try:
        del final_values[key]
        with open(filename, "wb") as outfile:
            data = serialize_to_json(final_values)
            outfile.write(data.encode("utf-8"))
    except KeyError:
        # If they key didn't exist then there is nothing to remove.
        pass


def record_deployed_values(deployed_values: Dict[str, Any], filename: str) -> None:
    """Record deployed values to a JSON file.

    This allows subsequent deploys to lookup previously deployed values.

    """
    final_values: Dict[str, Any] = {}
    if os.path.isfile(filename):
        with open(filename, "r") as f:
            final_values = json.load(f)
    final_values.update(deployed_values)
    with open(filename, "wb") as outfile:
        data = serialize_to_json(final_values)
        outfile.write(data.encode("utf-8"))


def serialize_to_json(data: Any) -> str:
    """Serialize to pretty printed JSON.

    This includes using 2 space indentation, no trailing whitespace, and
    including a newline at the end of the JSON document.  Useful when you want
    to serialize JSON to disk.

    """
    return json.dumps(data, indent=2, separators=(",", ": ")) + "\n"


class ChaliceZipFile(zipfile.ZipFile):
    """Support deterministic zipfile generation.

    Normalizes datetime and permissions.

    """

    compression = 0  # Try to make mypy happy.
    _default_time_time = (1980, 1, 1, 0, 0, 0)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._osutils = cast(OSUtils, kwargs.pop("osutils", OSUtils()))
        super(ChaliceZipFile, self).__init__(*args, **kwargs)

    # pylint: disable=W0221
    def write(
        self,
        filename: StrPath,
        arcname: Optional[StrPath] = None,
        compress_type: OptInt = None,
        compresslevel: OptInt = None,
    ) -> None:
        # Only supports files, py2.7 and 3 have different signatures.
        # We know that in our packager code we never call write() on
        # directories.
        zinfo = self._create_zipinfo(filename, arcname, compress_type)
        with open(filename, "rb") as f:
            self.writestr(zinfo, f.read())

    def _create_zipinfo(
        self,
        filename: StrPath,
        arcname: Optional[StrPath],
        compress_type: Optional[int],
    ) -> zipfile.ZipInfo:
        # The main thing that prevents deterministic zip file generation
        # is that the mtime of the file is included in the zip metadata.
        # We don't actually care what the mtime is when we run on lambda,
        # so we always set it to the default value (which comes from
        # zipfile.py).  This ensures that as long as the file contents don't
        # change (or the permissions) then we'll always generate the exact
        # same zip file bytes.
        # We also can't use ZipInfo.from_file(), it's only in python3.
        st = self._osutils.stat(str(filename))
        if arcname is None:
            arcname = filename
        arcname = self._osutils.normalized_filename(str(arcname))
        arcname = arcname.lstrip(os.sep)
        zinfo = zipfile.ZipInfo(arcname, self._default_time_time)
        # The external_attr needs the upper 16 bits to be the file mode
        # so we have to shift it up to the right place.
        zinfo.external_attr = (st.st_mode & 0xFFFF) << 16
        zinfo.file_size = st.st_size
        zinfo.compress_type = compress_type or self.compression
        return zinfo


def create_zip_file(source_dir: str, outfile: str) -> None:
    """Create a zip file from a source input directory.

    This function is intended to be an equivalent to
    `zip -r`.  You give it a source directory, `source_dir`,
    and it will recursively zip up the files into a zipfile
    specified by the `outfile` argument.

    """
    with ChaliceZipFile(
        outfile, "w", compression=zipfile.ZIP_DEFLATED, osutils=OSUtils()
    ) as z:
        for root, _, filenames in os.walk(source_dir):
            for filename in filenames:
                full_name = os.path.join(root, filename)
                archive_name = os.path.relpath(full_name, source_dir)
                z.write(full_name, archive_name)


class OSUtils(object):
    ZIP_DEFLATED = zipfile.ZIP_DEFLATED

    def environ(self) -> MutableMapping:
        return os.environ

    def open(self, filename: str, mode: str) -> IO:
        return open(filename, mode)

    def open_zip(
        self, filename: str, mode: str, compression: int = ZIP_DEFLATED
    ) -> zipfile.ZipFile:
        return ChaliceZipFile(filename, mode, compression=compression, osutils=self)

    def remove_file(self, filename: str) -> None:
        """Remove a file, noop if file does not exist."""
        # Unlike os.remove, if the file does not exist,
        # then this method does nothing.
        try:
            os.remove(filename)
        except OSError:
            pass

    def file_exists(self, filename: str) -> bool:
        return os.path.isfile(filename)

    def get_file_contents(
        self, filename: str, binary: bool = True, encoding: Any = "utf-8"
    ) -> str:
        # It looks like the type definition for io.open is wrong.
        # the encoding arg is unicode, but the actual type is
        # Optional[Text].  For now we have to use Any to keep mypy happy.
        if binary:
            mode = "rb"
            # In binary mode the encoding is not used and most be None.
            encoding = None
        else:
            mode = "r"
        with io.open(filename, mode, encoding=encoding) as f:
            return f.read()

    def set_file_contents(
        self, filename: str, contents: str, binary: bool = True
    ) -> None:
        if binary:
            mode = "wb"
        else:
            mode = "w"
        with open(filename, mode) as f:
            f.write(contents)

    def extract_zipfile(self, zipfile_path: str, unpack_dir: str) -> None:
        with zipfile.ZipFile(zipfile_path, "r") as z:
            z.extractall(unpack_dir)

    def extract_tarfile(self, tarfile_path: str, unpack_dir: str) -> None:
        with tarfile.open(tarfile_path, "r:*") as tar:
            tar.extractall(unpack_dir)

    def directory_exists(self, path: str) -> bool:
        return os.path.isdir(path)

    def get_directory_contents(self, path: str) -> List[str]:
        return os.listdir(path)

    def makedirs(self, path: str) -> None:
        os.makedirs(path)

    def dirname(self, path: str) -> str:
        return os.path.dirname(path)

    def abspath(self, path: str) -> str:
        return os.path.abspath(path)

    def joinpath(self, *args: str) -> str:
        return os.path.join(*args)

    def walk(
        self, path: str, followlinks: bool = False
    ) -> Iterator[Tuple[str, List[str], List[str]]]:
        return os.walk(path, followlinks=followlinks)

    def copytree(self, source: str, destination: str) -> None:
        if not os.path.exists(destination):
            self.makedirs(destination)
        names = self.get_directory_contents(source)
        for name in names:
            new_source = os.path.join(source, name)
            new_destination = os.path.join(destination, name)
            if os.path.isdir(new_source):
                self.copytree(new_source, new_destination)
            else:
                shutil.copy2(new_source, new_destination)

    def rmtree(self, directory: str) -> None:
        shutil.rmtree(directory)

    def copy(self, source: str, destination: str) -> None:
        shutil.copy(source, destination)

    def move(self, source: str, destination: str) -> None:
        shutil.move(source, destination)

    @contextlib.contextmanager
    def tempdir(self) -> Any:
        tempdir = tempfile.mkdtemp()
        try:
            yield tempdir
        finally:
            shutil.rmtree(tempdir)

    def popen(
        self,
        command: List[str],
        stdout: OptInt = None,
        stderr: OptInt = None,
        env: Optional[EnvVars] = None,
    ) -> subprocess.Popen:
        p = subprocess.Popen(command, stdout=stdout, stderr=stderr, env=env)
        return p

    def mtime(self, path: str) -> float:
        return os.stat(path).st_mtime

    def stat(self, path: str) -> os.stat_result:
        return os.stat(path)

    def normalized_filename(self, path: str) -> str:
        """Normalize a path into a filename.

        This will normalize a file and remove any 'drive' component
        from the path on OSes that support drive specifications.

        """
        return os.path.normpath(os.path.splitdrive(path)[1])

    @property
    def pipe(self) -> int:
        return subprocess.PIPE

    def basename(self, path: str) -> str:
        return os.path.basename(path)


def getting_started_prompt(prompter: Any) -> bool:
    return prompter.prompt(WELCOME_PROMPT)


class UI(object):
    def __init__(
        self,
        out: Optional[IO] = None,
        err: Optional[IO] = None,
        confirm: Optional[Any] = None,
    ) -> None:
        # I tried using a more exact type for the 'confirm'
        # param, but mypy seems to miss the 'if confirm is None'
        # check and types _confirm as Union[..., None].
        # So for now, we're using Any for this type.
        if out is None:
            out = sys.stdout
        if err is None:
            err = sys.stderr
        if confirm is None:
            confirm = click.confirm
        self._out = out
        self._err = err
        self._confirm = confirm

    def write(self, msg: str) -> None:
        self._out.write(msg)

    def error(self, msg: str) -> None:
        self._err.write(msg)

    def confirm(self, msg: str, default: bool = False, abort: bool = False) -> Any:
        try:
            return self._confirm(msg, default, abort)
        except click.Abort:
            raise AbortedError()


class PipeReader(object):
    def __init__(self, stream: IO[bytes]) -> None:
        self._stream = stream

    def read(self) -> OptBytes:
        if not self._stream.isatty():
            return self._stream.read()
        return None


class TimestampConverter(object):
    # This is roughly based off of what's used in the AWS CLI.
    _RELATIVE_TIMESTAMP_REGEX = re.compile(r"(?P<amount>\d+)(?P<unit>s|m|h|d|w)$")
    _TO_SECONDS = {
        "s": 1,
        "m": 60,
        "h": 3600,
        "d": 24 * 3600,
        "w": 7 * 24 * 3600,
    }

    def __init__(self, now: Optional[Callable[[], datetime]] = None) -> None:
        if now is None:
            now = datetime.utcnow
        self._now = now

    def timestamp_to_datetime(self, timestamp: str) -> datetime:
        """Convert a timestamp to a datetime object.

        This method detects what type of timestamp is provided and
        parse is appropriately to a timestamp object.

        """
        re_match = self._RELATIVE_TIMESTAMP_REGEX.match(timestamp)
        if re_match:
            datetime_value = self._relative_timestamp_to_datetime(
                int(re_match.group("amount")), re_match.group("unit")
            )
        else:
            datetime_value = self.parse_iso8601_timestamp(timestamp)
        return datetime_value

    def _relative_timestamp_to_datetime(self, amount: int, unit: str) -> datetime:
        multiplier = self._TO_SECONDS[unit]
        return self._now() + timedelta(seconds=amount * multiplier * -1)

    def parse_iso8601_timestamp(self, timestamp: str) -> datetime:
        return dateutil.parser.parse(timestamp, tzinfos={"GMT": tzutc()})


def get_current_runtime() -> str:
    return "python%s.%s" % (sys.version_info[0], sys.version_info[1])



import json
import os
from typing import Any, Dict
from attr import define

from hangar_sdk.constructs import S3, Asset, FilePath, LambdaFunction, LambdaLayer
from hangar_sdk.library import CompositeResource, HangarScope
from hangar_sdk.serverless.packager import (
    AppOnlyDeploymentPackager,
    LayerDeploymentPackager,
)
from hangar_sdk.serverless.utils import get_current_runtime

from hangar_sdk.config import Config
import boto3


@define(frozen=False, slots=False)
class HangarLambdaPackager(CompositeResource):
    dir_path: str
    runtime: str
    layer: str = ""
    only_app: str = ""

    def _resolve(self):
        from hangar_sdk.serverless.utils import UI, OSUtils

        from hangar_sdk.serverless.packager import DependencyBuilder

        dep_builder = DependencyBuilder(
            osutils=OSUtils(),
        )

        # packager = LambdaDeploymentPackager(
        #     osutils=OSUtils(),
        #     dependency_builder=dep_builder,
        #     ui=UI()
        # )

        layer_packager = LayerDeploymentPackager(
            osutils=OSUtils(), dependency_builder=dep_builder, ui=UI()
        )

        app_packager = AppOnlyDeploymentPackager(
            osutils=OSUtils(), dependency_builder=dep_builder, ui=UI()
        )

        # filename = packager.create_deployment_package(
        #     project_dir=self.dir_path,
        #     python_version=self.runtime
        # )

        layer = layer_packager.create_deployment_package(
            project_dir=self.dir_path, python_version=self.runtime
        )

        only_app = app_packager.create_deployment_package(
            project_dir=self.dir_path, python_version=self.runtime
        )

        # self.app = filename
        self.layer = layer
        self.only_app = only_app


class HangarServerlessFunction:
    def __init__(self, func, scope: HangarScope, secrets: Dict[str, str] = {}):
        self.func = func
        self.scope = scope
        self.secrets = secrets

    def __call__(self, event, context) -> Any:
        args = event.get("args", [])
        kwargs = event.get("kwargs", {})

        print("test")

        for key, value in self.secrets.items():
            print(boto3.client("ssm").get_parameter(Name=value, WithDecryption=True))
            value = boto3.client("ssm").get_parameter(Name=value, WithDecryption=True)[
                "Parameter"
            ]["Value"]
            os.environ[key] = value
            print(f"Setting {key} to {value}")

        return self.func(*args, **kwargs)

    def invoke(self, *args, **kwargs):
        payload = {"args": args, "kwargs": kwargs}

        client = self.scope.get_boto3_session().client("lambda")

        response = client.invoke(
            FunctionName="hangar_serverless_" + self.func.__name__,
            InvocationType="RequestResponse",
            Payload=json.dumps(payload),
        )
        return json.loads(response["Payload"].read().decode("utf-8"))


class LocalHangarServerlessFunction:
    def __init__(self, func, scope) -> None:
        self.func = func
        self.scope = scope

    def __call__(self, *args, **kwargs) -> Any:
        if os.getenv("HANGAR_ENVIRONMENT") == "LOCAL":
            payload = {"args": args, "kwargs": kwargs}
            client = self.scope.get_boto3_session().client("lambda")

            response = client.invoke(
                FunctionName="hangar_serverless_" + self.func.__name__,
                InvocationType="RequestResponse",
                Payload=json.dumps(payload),
            )
            return json.loads(response["Payload"].read().decode("utf-8"))

        else:
            raise Exception(
                "Cannot call remote function. Call within a local function."
            )

    def invoke(self, *args, **kwargs):
        return self(*args, **kwargs)


class Serverless:
    def __init__(self, name, path, scope: HangarScope, bucket: S3, role_arn: str):
        self.name = name
        self._functions = []
        self.file = path
        self.local_entrypoint = None
        self.scope = scope
        self.bucket = bucket
        self.role_arn = role_arn
        self.mode = Config.mode

    @property
    def env(self):
        return os.getenv("HANGAR_ENVIRONMENT")

    def Function(self, timeout=60, environment_variables={}, role=None, secrets={}):
        def function_decorator(func):
            if self.env == "REMOTE":
                print("Remote")
                return HangarServerlessFunction(func, self.scope, secrets=secrets)

            if self.env == "LOCAL":
                return HangarServerlessFunction(func, self.scope, secrets=secrets)
            else:
                functionPath = (
                    os.path.basename(self.file).split(".")[0] + "." + func.__name__
                )
                self._functions.append(
                    {
                        "path": functionPath,
                        "timeout": timeout,
                        "environment_variables": environment_variables,
                        "role": role,
                    }
                )
                return LocalHangarServerlessFunction(func, self.scope)

        return function_decorator

    def LocalEntrypoint(self, func):
        self.local_function = func
        return

    async def _resolve(self):
        if self.env == "REMOTE":
            return

        if self.mode == "invoke":
            if self.local_function:
                os.environ["HANGAR_ENVIRONMENT"] = "LOCAL"
                self.local_function()
                del os.environ["HANGAR_ENVIRONMENT"]

            return

        current_runtime = get_current_runtime()

        packager = HangarLambdaPackager(
            dir_path=os.path.dirname(self.file),
            runtime=current_runtime,
            name="hangar_serverless_packager",
            scope=HangarScope(
                name="temp",
                region=self.scope.region,
                roleAssumptions=self.scope.roleAssumptions,
                accountName=self.scope.accountName,
                endpoint=self.scope.endpoint,
                api_key=self.scope.api_key,

            )
        )

        packager._resolve()

        self._resources = []

        asset_bucket = self.bucket

        layer_asset = Asset(
            name=self.name + "-hangar_serverless_layer_asset",
            source=FilePath(
                name=os.path.basename(packager.layer),
                path=packager.layer,
                scope=self.scope,
            ),
            bucket=asset_bucket,
            scope=self.scope,
        )

        app_asset = Asset(
            name=self.name + "-hangar-serverless-app-asset",
            source=FilePath(
                name=os.path.basename(packager.only_app),
                path=packager.only_app,
                scope=self.scope,
            ),
            bucket=asset_bucket,
            scope=self.scope,
        )
        # app_asset.sync()

        layer = LambdaLayer(
            scope=self.scope,
            name=self.name + "-hangar_serverless_layer",
            asset=layer_asset,
            runtimes=[current_runtime],
        )

        for func in self._functions:
            LambdaFunction(
                scope=self.scope,
                name="hangar_serverless_" + func["path"].split(".")[1],
                handler=func["path"],
                asset=app_asset,
                function_name=func["path"].split(".")[1],
                role=self.role_arn if func["role"] is None else func["role"],
                runtime=current_runtime,
                timeout=func["timeout"],
                environment={
                    "HANGAR_ENVIRONMENT": "REMOTE",
                    "HANGAR_API_KEY": self.scope.api_key,
                    **func["environment_variables"],
                },
                layers=[layer],
            )

        print("Deploying")

        await self.scope.deploy()

        return self._resources



import inspect
import os
from typing import Annotated, Optional

import httpx
import typer
from rich.console import Console
from rich.table import Table
from rich.tree import Tree

from hangar_sdk.config import Config

# commands

# actions
# init : creates a new project
# up : runs python and deploys resources
# down : destroys resources


# traits
# logs : shows logs
# ssh : ssh into instances

# CRUD

# resources :
#  - list resources
#  - describe resource
#  - state of resource
#  - last applied and status

# scope : shows scope
# - list scopes
# - describe scope
# - list resources
def version():
    import importlib.metadata
    return importlib.metadata.version("hangar-sdk")

def print_tree(data, parent=None, tree=None):
    # Initialize the tree if not provided
    if tree is None:
        tree = Tree("Root")
        parent = tree

    if parent is None:
        parent = tree

    # Iterate over items in the dictionary
    for key, value in data.items():
        # For dictionary, recursively build the tree
        if isinstance(value, dict):
            branch = parent.add(f"{key}: ")
            print_tree(value, branch, tree)
        # For list, add each item as a separate branch
        elif isinstance(value, list):
            branch = parent.add(f"{key}: [")
            for item in value:
                if isinstance(item, dict):
                    print_tree(item, branch, tree)
                else:
                    branch.add(str(item))
            branch.add("]")
        # For other types, just add them to the current branch
        else:
            parent.add(f"{key}: {value}")

    return tree


app = typer.Typer(pretty_exceptions_enable=False)

__version__ = "0.1.0"

def version_callback(value: bool):
    if value:
        print(f"Hangar CLI: {version()}")
        raise typer.Exit()

@app.callback()
def main(version: Annotated[
        Optional[bool], typer.Option("--version", callback=version_callback)
    ] = None,):
    """
    A CLI application using Typer.
    """
    # print("Hangar CLI",version)
    pass


resources = typer.Typer()

app.add_typer(resources, name="resources")

# # define type URL OPTION

UrlOption = Annotated[
    Optional[str], typer.Option("--url", help="Hangar URL", envvar="HANGAR_URL")
]
DEFAULT_URL = "https://api.tryhangar.com/controlplane"

ApiKeyOption = Annotated[str, typer.Option(help="Api key", envvar="HANGAR_API_KEY")]
JsonOption = Annotated[Optional[bool], typer.Option("--json", help="Output JSON")]

HANGAR_MAIN_URL = "app.tryhangar.com"


@resources.command(name="list")
def list_resources(
    api_key: ApiKeyOption,
    json: JsonOption = False,
    url: UrlOption = DEFAULT_URL,
):
    r = httpx.request(
        "GET",
        url + "/resources",
        headers={"X-API-KEY": api_key},
    )

    if json:
        print(r.json())
        return

    table = Table(title="Resources")

    table.add_column("Resource Id")

    for resource in r.json():
        table.add_row(resource["ResourceId"])

    console = Console()
    console.print(table)


@resources.command()
def describe(
    resource_id: str,
    api_key: ApiKeyOption,
    json: JsonOption = False,
    url: UrlOption = DEFAULT_URL,
):
    r = httpx.request(
        "GET",
        f"{url}/resources/{resource_id}",
        headers={"X-API-KEY": api_key},
    )
    # print(r.json())
    resource = r.json()

    if json:
        print(resource)
        return

    tree = Table(title="Resource: " + resource_id)
    tree.add_column("Scope")
    tree.add_column("Errors")
    tree.add_column("Last Applied")
    tree.add_column("RoleAssumptions")
    tree.add_column("AccountId")
    tree.add_column("Region")
    tree.add_column("Version")

    scope = resource.get("Scope", "")
    error_str = ""
    if "ControlStatus" in resource:
        errors = resource["ControlStatus"]["status"]["errors"]
        for error in errors:
            error_str += error["resource_id"] + ":" + error["message"] + "\n"
    else:
        error_str = "No Errors :)"

    last_applied = resource.get("LastApplied", "")

    role_assumptions = str(resource.get("RoleAssumptions", ""))
    if role_assumptions == "[]":
        role_assumptions = "Default Role"

    account_id = resource.get("AccountId", "")
    region = resource.get("Region", "")
    version = resource.get("Version", "")

    tree.add_row(
        scope, error_str, last_applied, role_assumptions, account_id, region, version
    )

    console = Console()
    console.print(tree)


@resources.command()
def state(
    resource_id: str,
    api_key: ApiKeyOption,
    json: JsonOption = False,
    url: UrlOption = DEFAULT_URL,
):
    r = httpx.request(
        "GET",
        f"{url}/resources/{resource_id}",
        headers={"X-API-KEY": api_key},
        params={"resourceId": resource_id},
    )

    resource = r.json()
    if json:
        print(resource)
        return
    state = resource["ControlState"]

    tree = Tree("State for: " + resource_id)

    for r, v in state.items():
        resource = tree.add(r)
        print_tree(v["values"], tree=resource)

    console = Console()
    console.print(tree)


scopes = typer.Typer()

app.add_typer(scopes, name="scopes")


@scopes.command(name="list")
def list_scopes(
    api_key: ApiKeyOption,
    url: UrlOption = DEFAULT_URL,
):
    r = httpx.request(
        "GET",
        f"{url}/scopes",
        headers={"X-API-KEY": api_key},
    )
    # print(r.json())
    table = Table(title="Scopes")

    table.add_column("Scope Id")

    for scope in r.json():
        table.add_row(scope)

    console = Console()
    console.print(table)


@scopes.command(name="describe")
def describe_scope(
    scope_id: str,
    api_key: ApiKeyOption,
    url: UrlOption = DEFAULT_URL,
):
    r = httpx.request(
        "GET",
        f"{url}/scopes/" + scope_id,
        headers={"X-API-KEY": api_key},
    )
    res = r.json()[0]

    table = Table(title="Scope: " + scope_id)
    table.add_column("Scope")
    table.add_column("Roles")
    table.add_column("AwsAccount")
    table.add_column("Region")

    scope = res.get("Scope", "")
    role_assumptions = str(res.get("RoleAssumptions", ""))

    if role_assumptions == "[]":
        role_assumptions = "Default Role"

    account_id = res.get("AccountId", "")
    region = res.get("Region", "")

    table.add_row(scope, role_assumptions, account_id, region)

    console = Console()
    console.print(table)


@scopes.command(name="resources")
def list_scope_resources(
    scope: str, api_key: ApiKeyOption, url: UrlOption = DEFAULT_URL
):
    r = httpx.request(
        "GET",
        f"{url}/scopes/{scope}/resources",
        headers={"X-API-KEY": api_key},
    )

    table = Table(title="Scope: " + scope)

    table.add_column("Resource Id")

    for resource in r.json():
        table.add_row(resource["ResourceId"])

    console = Console()
    console.print(table)


logs = typer.Typer()

app.add_typer(logs, name="logs")


@logs.command(name="list")
def list_logs(api_key: ApiKeyOption, url: UrlOption = DEFAULT_URL):
    r = httpx.request(
        "GET",
        f"{url}/resources",
        headers={"X-API-KEY": api_key},
    )

    resource_list = r.json()

    table = Table()
    table.add_column("Scope")
    table.add_column("ResourceId")
    table.add_column("Log identifier")

    for resource in resource_list:
        # print(resource["Config"])
        if (
            resource["Config"]
            and resource["Config"] != {}
            and resource["Config"]["construct"] == "cluster"
            # and resource["Config"]["services"] != []
        ):
            # print("hello")
            for service in resource["Config"]["services"]:
                # print(resource["Region"], resource["Config"]["name"], service["name"])
                table.add_row(
                    resource["Scope"],
                    resource["ResourceId"],
                    resource["Region"]
                    + ":"
                    + resource["Config"]["name"]
                    + ":"
                    + service["name"]
                    + "/task",
                )

    console = Console()

    console.print(table)


# @logs.command("tail")
# def tail(
#     log_group: str,
#     api_key: Annotated[str, typer.Option(help="Api key", envvar="HANGAR_API_KEY")],
# ):
#     with httpx.stream(
#         "GET",
#         "http://localhost:8000/controlplane/logs",
#         headers={"X-API-KEY": api_key},
#         params={"log_group_name": log_group},
#     ) as r:
#         for line in r.iter_lines():
#             print(line)


@app.command()
def up(
    file: str, api_key: ApiKeyOption, url: UrlOption = DEFAULT_URL, mode: str = "create"
):
    typer.echo("Deploying resources...")
    import importlib.util

    os.environ["HANGAR_API_KEY"] = api_key
    os.environ["HANGAR_URL"] = url
    Config.mode = mode
    spec = importlib.util.spec_from_file_location("hangar_module", file)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    if hasattr(mod, "hangar_run"):
        if inspect.iscoroutinefunction(mod.hangar_run):
            import asyncio

            asyncio.run(mod.hangar_run())
        else:
            mod.hangar_run()


@app.command()
def down(file: str, api_key: ApiKeyOption, url: UrlOption = DEFAULT_URL):
    typer.echo("Destroying resources...")
    import importlib.util

    from hangar_sdk.library import Config

    Config.mode = "delete"

    spec = importlib.util.spec_from_file_location("hangar_module", file)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    os.environ["HANGAR_API_KEY"] = api_key
    os.environ["HANGAR_URL"] = url

    if hasattr(mod, "hangar_run"):
        if inspect.iscoroutinefunction(mod.hangar_run):
            import asyncio

            asyncio.run(mod.hangar_run())
        else:
            mod.hangar_run()


@app.command()
def init():
    typer.echo("Initializing project...")
    api_key = typer.prompt("Api key: ")
    url = typer.prompt("Hangar URL: ")

    # set env vars
    os.environ["HANGAR_API_KEY"] = api_key
    os.environ["HANGAR_URL"] = url









import os
import pathlib
import tarfile
from abc import ABC, abstractmethod
from typing import Dict, List, Literal, Optional, Union

from aiohttp import ClientError
from attr import define
from pydantic import BaseModel

from .base import HangarScope
from .library import CompositeResource, Deployable


class SourceInterface(ABC):
    name: str

    @abstractmethod
    def __init__(self, source):
        pass


@define(kw_only=True, slots=False)
class S3(CompositeResource, Deployable):
    @property
    def client(self):
        session = self.scope.get_boto3_session()
        return session.client("s3")

    def create_construct_definition(self) -> dict:
        return {
            "construct": "s3",
            "name": self.name,
        }


@define(kw_only=True, slots=False)
class GCS(CompositeResource, Deployable):
    location: str

    def create_construct_definition(self) -> dict:
        return {"construct": "gcs", "name": self.name, "location": self.location}


@define(kw_only=True, slots=False)
class Asset(CompositeResource, Deployable):
    source: SourceInterface
    bucket: S3

    def create_construct_definition(self) -> dict:
        return {
            "construct": "asset",
            "name": self.name,
            "source": {"!REF": True, "resourceId": self.source.name},
            "bucket": self.bucket._get_ref(),
        }

    def post_resolve(self):
        print("uploading asset")
        self.sync()

    def sync(self, token=None):
        if self.mode == "delete":
            return

        if isinstance(self.source, DirPath):
            fname = f"{self.source.name}.tar"

            ignorefile = (
                pathlib.Path(self.source.ignorefile) if self.source.ignorefile else None
            )

            ignorefile = (
                pathlib.Path(self.source.ignorefile) if self.source.ignorefile else None
            )
            if ignorefile and ignorefile.exists():
                from gitignore_parser import parse_gitignore

                matches = parse_gitignore(ignorefile.as_posix())
                with tarfile.open(fname, "w") as tar:
                    tar.add(
                        self.source.path,
                        arcname=os.path.basename(self.source.path),
                        filter=lambda x: x if not matches(x.name) else None,
                    )
            else:
                with tarfile.open(fname, "w") as tar:
                    tar.add(
                        self.source.path, arcname=os.path.basename(self.source.path)
                    )

            try:
                self.bucket.client.upload_file(
                    fname, self.bucket.name, f"{self.source.name}/{fname}"
                )
            except ClientError as e:
                print(e)

        elif isinstance(self.source, FilePath):
            try:
                self.bucket.client.upload_file(
                    self.source.path,
                    self.bucket.name,
                    self.source.name
                    if self.source.name
                    else os.path.basename(self.source.path),
                )
            except ClientError as e:
                print(e)

        elif isinstance(self.source, GitHubRepo):
            self.scope.execute_action(
                self.name, "sync", {"token": token} if token is not None else {}
            )


@define(kw_only=True, slots=False)
class DirPath(CompositeResource, SourceInterface, Deployable):
    path: str
    name: str
    ignorefile: Union[str, None] = None

    def create_construct_definition(self) -> dict:
        return {
            "construct": "dirpath",
            "name": self.name,
            "path": self.path,
        }


@define(kw_only=True, slots=False)
class FilePath(CompositeResource, SourceInterface, Deployable):
    path: str
    name: Optional[str]

    def create_construct_definition(self) -> dict:
        return {
            "construct": "filepath",
            "name": self.name,
            "path": self.path,
        }


@define(kw_only=True, slots=False)
class GitHubRepo(CompositeResource, SourceInterface, Deployable):
    repo: str
    branch: str

    def create_construct_definition(self) -> dict:
        return {
            "construct": "githubrepo",
            "name": self.name,
            "repository": self.repo,
            "branch": self.branch,
        }


class BuilderInterface(ABC):
    builderType: str

    @abstractmethod
    def __init__(self, builder):
        pass


@define(kw_only=True, slots=False)
class BuildkitBuilder(CompositeResource, BuilderInterface, Deployable):
    builder_type = "buildkit"

    def create_construct_definition(self) -> dict:
        return {"construct": "buildkit", "name": self.name}


class RegistryInterface(ABC):
    name: str

    @abstractmethod
    def __init__(self, builder):
        pass


@define(kw_only=True, slots=False)
class Registry(CompositeResource, RegistryInterface, Deployable):
    def create_construct_definition(self):
        return {"construct": "ecr"}


@define(kw_only=True, slots=False)
class ContainerBuilder(CompositeResource, Deployable):
    asset: Asset
    builder: BuilderInterface
    registry: RegistryInterface

    def create_construct_definition(self):
        return {
            "construct": "codebuild",
            "asset": self.asset._get_ref(),
            "builder": self.builder._get_ref(),
            "registry": self.registry._get_ref(),
        }

    def build(self, build_context=None, tag="latest"):
        if self.mode == "delete":
            return
        session = self.scope.get_boto3_session()
        client = session.client("codebuild")
        env_variables = [
            {
                "name": "BUILD_CONTEXT",
                "value": build_context,
                "type": "PLAINTEXT",
            },
            {
                "name": "TAG",
                "value": tag,
                "type": "PLAINTEXT",
            },
        ]
        # Start a build with environment variable overrides
        return client.start_build(
            projectName=self.name,
            environmentVariablesOverride=env_variables,
        )

        # return self.scope.execute_action(
        #     self.name,
        #     "build",
        #     {
        #         "build_context": build_context
        #         if build_context is not None
        #         else self.name,
        #         "tag": tag
        #     },
        # )

    def poll(self, build_id=None):
        if self.mode == "delete":
            return

        if build_id is None:
            print("no build passed, skipping polling")
            return

        session = self.scope.get_boto3_session()
        codebuild = session.client("codebuild")
        build = codebuild.batch_get_builds(ids=[build_id])
        return build["builds"][0]


class PortMappings(BaseModel):
    containerPort: int
    hostPort: int


@define(kw_only=True, slots=False)
class Container(CompositeResource, Deployable):
    source: Registry
    tag: str
    startCommand: str = None
    portMappings: List[PortMappings] = None

    def create_construct_definition(self) -> dict:
        return {
            "construct": "container",
            "registry": self.source._get_ref(),
            "tag": self.tag,
            "startCommand": self.startCommand.split(" ")
            if self.startCommand is not None
            else None,
            "portMappings": [mapping.model_dump() for mapping in self.portMappings]
            if self.portMappings is not None
            else None,
        }


@define(kw_only=True, slots=False)
class Service(CompositeResource):
    name: str
    container: Container
    environmentVariables: Union[Dict[str, str], None] = None


def env_vars_dict_to_str(env_vars: Dict[str, str]):
    return "\n".join([f"{k}={v}" for k, v in env_vars.items()])


@define(kw_only=True, slots=False)
class Cluster(CompositeResource, Deployable):
    name: str
    services: List[Service]

    def log(self, path=""):
        if self.mode == "delete":
            return
        return self.scope.get_logs(self.name, path)

    def create_construct_definition(self) -> dict:
        return {
            "construct": "cluster",
            "name": self.name,
            "services": [
                {
                    "name": s.name,
                    "container": s.container._get_ref(),
                    "environment_variables": env_vars_dict_to_str(
                        s.environmentVariables
                    )
                    if s.environmentVariables
                    else "",
                }
                for s in self.services
            ],
        }


class HostInterface(ABC):
    name: str

    @abstractmethod
    def __init__(self, builder):
        pass


class Ingress(BaseModel):
    fromPort: int
    toPort: int
    protocol: str
    cidrIp: List[str]


class Egress(BaseModel):
    fromPort: int
    toPort: int
    protocol: str
    cidrIp: List[str]


@define(kw_only=True, slots=False)
class SecurityGroup(CompositeResource, Deployable):
    name: str
    ingresses: List[Ingress] = None
    egresses: List[Egress] = None

    def create_construct_definition(self) -> dict:
        return {
            "construct": "securitygroup",
            "ingresses": [
                {
                    "fromPort": str(ingress.fromPort),
                    "toPort": str(ingress.toPort),
                    "protocol": ingress.protocol,
                    "cidrIp": ingress.cidrIp,
                }
                for ingress in self.ingresses
            ],
            "egresses": [
                {
                    "fromPort": str(egress.fromPort),
                    "toPort": str(egress.toPort),
                    "protocol": egress.protocol,
                    "cidrIp": egress.cidrIp,
                }
                for egress in self.egresses
            ],
        }


@define(kw_only=True, slots=False)
class Instance(CompositeResource, Deployable):
    name: str
    security_groups: List[SecurityGroup]
    instance_type: str
    path_to_key: Union[str, None] = None
    ssh: bool = True

    def pre_resolve(self):
        if self.ssh:
            works = False
            for sg in self.security_groups:
                if (
                    Ingress(
                        protocol="tcp",
                        fromPort=22,
                        toPort=22,
                        cidrIp=["0.0.0.0/0"],
                    )
                    in sg.ingresses
                ):
                    works = True
                    break
            if not works:
                self.security_groups[0].ingresses.append(
                    Ingress(
                        protocol="tcp",
                        fromPort=22,
                        toPort=22,
                        cidrIp=["0.0.0.0/0"],
                    )
                )

    def create_construct_definition(self) -> dict:
        return {
            "construct": "instance",
            "securityGroups": [sg._get_ref() for sg in self.security_groups],
            "instance_type": self.instance_type,
        }

    def run_command(self, command: str):
        if self.mode == "delete":
            return
        import os
        import sys

        import paramiko

        # Function to get default SSH key path based on the operating system
        def get_default_ssh_key_path():
            home = os.path.expanduser("~")
            if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
                # Default path for Linux and macOS
                return os.path.join(home, ".ssh", "id_rsa")
            elif sys.platform.startswith("win32"):
                # Default path for Windows
                return os.path.join(home, ".ssh", "id_rsa")
            else:
                raise ValueError("Unsupported operating system")

        INSTANCE_USER = "ubuntu"  # e.g., 'ec2-user' or 'ubuntu'

        if not self.path_to_key:
            # Prompt user for SSH key path or use default
            self.path_to_key = get_default_ssh_key_path()

        # Read your public key
        public_key_path = self.path_to_key + ".pub"
        with open(public_key_path, "r") as key_file:
            public_key = key_file.read().strip()

        try:
            self.scope.execute_action(
                self.name, "send_public_key", {"public_key": public_key}
            )
            print("Public key sent successfully.")
        except Exception as e:
            print(f"Failed to send public key: {e}")

        # Connect to the instance using Paramiko
        key = paramiko.RSAKey.from_private_key_file(self.path_to_key)
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            ssh_client.connect(
                hostname=self.state[self.name]["values"]["public_ip"],
                username=INSTANCE_USER,
                pkey=key,
            )
            print("SSH connection established.")

            _stdin, _stdout, _stderr = ssh_client.exec_command(command)
            _stdin.close()

        except Exception as e:
            print(self.state[self.name]["values"])
            print(f"SSH connection failed: {e}")
        finally:
            ssh_client.close()
            print("SSH connection closed.")

        # return command_results

        return {
            "command": command,
            "stdout": _stdout.read().decode(),
            "stderr": _stderr.read().decode(),
        }

    def run_commands(self, commands: List[str]):
        if self.mode == "delete":
            return

        import os
        import sys

        import paramiko

        # Function to get default SSH key path based on the operating system
        def get_default_ssh_key_path():
            home = os.path.expanduser("~")
            if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
                # Default path for Linux and macOS
                return os.path.join(home, ".ssh", "id_rsa")
            elif sys.platform.startswith("win32"):
                # Default path for Windows
                return os.path.join(home, ".ssh", "id_rsa")
            else:
                raise ValueError("Unsupported operating system")

        INSTANCE_USER = "ubuntu"  # e.g., 'ec2-user' or 'ubuntu'

        if not self.path_to_key:
            # Prompt user for SSH key path or use default
            self.path_to_key = get_default_ssh_key_path()

        # Read your public key
        public_key_path = self.path_to_key + ".pub"
        with open(public_key_path, "r") as key_file:
            public_key = key_file.read().strip()

        try:
            self.scope.execute_action(
                self.name, "send_public_key", {"public_key": public_key}
            )
            print("Public key sent successfully.")
        except Exception as e:
            print(f"Failed to send public key: {e}")

        # Connect to the instance using Paramiko
        key = paramiko.RSAKey.from_private_key_file(self.path_to_key)
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        try:
            ssh_client.connect(
                hostname=self.state[self.name]["values"]["public_ip"],
                username=INSTANCE_USER,
                pkey=key,
            )
            print("SSH connection established.")

            for command in commands:
                _stdin, _stdout, _stderr = ssh_client.exec_command(command)
                yield {
                    "command": command,
                    "stdout": _stdout.read().decode(),
                    "stderr": _stderr.read().decode(),
                }
                _stdin.close()

        except Exception as e:
            print(self.state[self.name]["values"])
            print(f"SSH connection failed: {e}")
        finally:
            ssh_client.close()
            print("SSH connection closed.")


class DockerEnvironment(CompositeResource):
    name: str
    host: Instance
    containers: List[Container]

    def post_resolve(self):
        self.host.run_commands(["sudo apt update", "sudo apt install -y docker.io"])
        self.ssh_host = self.host.state[self.host.name]["values"]["public_ip"]


@define(kw_only=True, slots=False)
class LambdaLayer(CompositeResource, Deployable):
    asset: Asset
    runtimes: Union[List[str], None] = None

    def create_construct_definition(self) -> dict:
        return {
            "construct": "lambda_layer",
            "name": self.name,
            "asset": self.asset._get_ref(),
            "runtimes": self.runtimes,
        }


@define(kw_only=True, slots=False)
class LambdaFunction(CompositeResource, Deployable):
    scope: HangarScope
    name: str
    function_name: str
    asset: Asset
    runtime: str
    handler: str
    timeout: int
    role: str
    environment: Union[Dict[str, str], None] = None
    layers: Union[List[LambdaLayer], None] = None

    def create_construct_definition(self):
        return {
            "construct": "lambda",
            "name": self.function_name,
            "asset": self.asset._get_ref(),
            "role": self.role,
            "runtime": self.runtime,
            "handler": self.handler,
            "timeout": self.timeout,
            "environment": self.environment,
            "layers": [layer._get_ref() for layer in self.layers]
            if self.layers
            else None,
        }


class Queue(CompositeResource, Deployable):
    name: str
    scope: HangarScope
    type: Union[Literal["fifo"], Literal["standard"]]

    def create_construct_definition(self) -> dict:
        return {
            "construct": "queue",
            "name": self.name,
            "config": {
                "type": self.type,
            },
        }

    def send_message(self, message: str):
        if self.mode == "delete":
            return
        self.scope.execute_action(self.name, "send_message", {"message": message})

    def receive_message(self):
        if self.mode == "delete":
            return
        return self.scope.execute_action(self.name, "receive_message", {})

    def delete_message(self, receipt_handle: str):
        if self.mode == "delete":
            return
        return self.scope.execute_action(
            self.name, "delete_message", {"receipt_handle": receipt_handle}
        )

    def purge_queue(self):
        if self.mode == "delete":
            return
        return self.scope.execute_action(self.name, "purge_queue", {})
